"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.enumerable = exports.empty = exports.Enumerable = exports.iterate = void 0;
function concat(iterators) {
    return {
        *[Symbol.iterator]() {
            for (let iterator of iterators) {
                for (let item of iterator) {
                    yield item;
                }
            }
        }
    };
}
function iterate(fn) {
    return {
        [Symbol.iterator]: fn
    };
}
exports.iterate = iterate;
function asGroup(key, source) {
    let result = enumerable(source);
    result.key = key;
    return result;
}
class Enumerable {
    constructor(items) {
        return enumerable(items);
    }
    static sum(items, selector) {
        let result = 0;
        for (let item of items) {
            result += selector(item);
        }
        return result;
    }
    static contains(items, element, selector = (pre, current) => pre === current) {
        for (let item of items) {
            if (selector(item, element)) {
                return true;
            }
        }
        return false;
    }
    static filter(items, predicate) {
        return iterate(function* () {
            for (let item of items) {
                if (predicate(item)) {
                    yield item;
                }
            }
        });
    }
    static repeat(element, count = 1, reset = -1) {
        let sureFunc = (typeof element === 'function' ? element : (index) => element);
        return enumerable(iterate(function* () {
            let canReset = reset > 0;
            for (let i = 0; i < count; i++)
                yield sureFunc(canReset ? ((i + 1) % reset) + 1 : i + 1);
        }));
    }
    static times(count, action) {
        for (let i = 0; i < count; i++) {
            action(count);
        }
    }
    static map(items, selector) {
        return iterate(function* () {
            for (let item of items) {
                yield selector(item);
            }
        });
    }
    static distinct(items, equalSelector = (pre, cu) => pre === cu) {
        return iterate(function* () {
            let array = Array.from(items);
            for (let i = 0; i < array.length; i++) {
                let indexOf = array.findIndex(a => equalSelector(array[i], a));
                if (indexOf === i) {
                    yield array[i];
                }
            }
        });
    }
    static reduce(items, selector) {
        let iterate = items[Symbol.iterator]();
        let { done, value } = iterate.next();
        if (done)
            throw new Error('empty items in reduce');
        let result = value;
        let hasNext = false;
        while (!hasNext) {
            let { done, value } = iterate.next();
            result = selector(result, value);
            hasNext = done;
        }
        return result;
    }
    static count(items) {
        if (Array.isArray(items))
            return items.length;
        let result = 0;
        for (let item of items) {
            result++;
        }
        return result;
    }
    static extract(items, extractItems, selector) {
        return iterate(function* () {
            for (let item of items) {
                let any = Enumerable.contains(extractItems, item, selector);
                if (!any) {
                    yield item;
                }
            }
        });
    }
    static take(items, value) {
        return iterate(function* () {
            let count = value;
            for (let item of items) {
                if (count <= 0)
                    break;
                yield item;
                count--;
            }
        });
    }
    static any(items, selector) {
        for (let item of items) {
            if (selector(item))
                return true;
        }
        return false;
    }
    static skip(items, value) {
        return iterate(function* () {
            let count = value;
            for (let item of items) {
                if (count <= 0) {
                    yield item;
                }
                else {
                    count--;
                }
            }
        });
    }
    static order(items, type, selector) {
        return Array.from(items).sort((a, b) => {
            let left = selector != null ? selector(a) : a;
            let right = selector != null ? selector(b) : b;
            if (typeof left === "string") {
                return type === 'asc' ? left.localeCompare(right) : right.localeCompare(left);
            }
            return type === 'asc' ? left - right : right - left;
        });
    }
    static forEach(items, action) {
        let index = 0;
        for (let item of items) {
            let isContinue = action(item, index);
            index++;
            if (isContinue == null || isContinue) {
                continue;
            }
            break;
        }
    }
    static concat(items) {
        return concat(items);
    }
    static group(items, selector, map) {
        let results = {};
        for (let item of items) {
            let s = selector(item);
            let key = createKey(s);
            let result = map(item);
            if (results[key] != null) {
                results[key].items.push(result);
            }
            else {
                results[key] = {
                    key: s,
                    items: [result]
                };
            }
        }
        return iterate(function* () {
            let keys = Object.keys(results);
            for (let i = 0; i < keys.length; i++) {
                let group = results[keys[i]];
                yield asGroup(group.key, group.items);
            }
        });
    }
    static lookup(items, parentSelector, selector, map) {
        let group = Array.from(Enumerable.group(items, parentSelector, (a) => a)).map((item, index) => ({ index, item }));
        return iterate(function* () {
            for (let g of group) {
                let id = createKey(g.item.key);
                let includes = group.filter(t => t.index != g.index && t.item.any(a => createKey(selector(a)) === id));
                for (let include of includes) {
                    let incGroup = group[include.index];
                    group[include.index].item = asGroup(incGroup.item.key, incGroup.item.concat(g.item));
                }
                ;
            }
            for (let o of group) {
                yield asGroup(o.item.key, o.item.map(t => map(t)));
            }
        });
    }
    static unwind(items, selector) {
        return iterate(function* () {
            for (let item of items) {
                let children = selector(item);
                for (let child of children) {
                    yield child;
                }
            }
        });
    }
    static unwindMap(items, selector, mapSelector) {
        return iterate(function* () {
            for (let item of items) {
                let children = selector(item);
                for (let child of children) {
                    yield mapSelector(child, item);
                }
            }
        });
    }
    static max(items, selector = (elem) => elem) {
        let max = NaN;
        for (let item of items) {
            let value = selector(item);
            max = (function () {
                if (max == null) {
                    return value;
                }
                return value > max ? value : max;
            })();
        }
        return max;
    }
    static min(items, selector = (elem) => elem) {
        let max = NaN;
        for (let item of items) {
            let value = selector(item);
            max = (function () {
                if (max == null) {
                    return value;
                }
                return value < max ? value : max;
            })();
        }
        return max;
    }
    static toObject(items, keySelector, map) {
        let o = {};
        for (let item of items) {
            o[keySelector(item)] = map(item);
        }
        return o;
    }
}
exports.Enumerable = Enumerable;
let createKey = function (key) {
    if (typeof key === "string") {
        return key;
    }
    if (typeof key === "number") {
        return key.toString();
    }
    if (key == null)
        return '';
    return JSON.stringify(key);
};
function empty() {
    return enumerable([]);
}
exports.empty = empty;
function enumerable(iterable) {
    return {
        get array() {
            return Array.from(this);
        },
        contains(element, selector = (pre, current) => pre === current) {
            return Enumerable.contains(this, element, selector);
        },
        sum(selector) {
            return Enumerable.sum(this, selector);
        },
        filter(predicate) {
            return enumerable(Enumerable.filter(this, predicate));
        },
        map(selector) {
            return enumerable(Enumerable.map(this, selector));
        },
        distinct(eqSelector = (pre, next) => pre == next) {
            return enumerable(Enumerable.distinct(this, eqSelector));
        },
        reduce(selector) {
            return Enumerable.reduce(this, selector);
        },
        any(selector) {
            return Enumerable.any(this, selector);
        },
        count() {
            return Enumerable.count(this);
        },
        extract(items, selector = (pre, next) => pre === next) {
            return enumerable(Enumerable.extract(this, items, selector));
        },
        take(value) {
            return enumerable(Enumerable.take(this, value));
        },
        skip(value) {
            return enumerable(Enumerable.skip(this, value));
        },
        order(type, selector = null) {
            if (type == null) {
                type = selector;
                selector = null;
            }
            return enumerable(Enumerable.order(this, type, selector));
        },
        forEach(action) {
            return Enumerable.forEach(this, action);
        },
        concat(...items) {
            let o = [this];
            items.forEach((item) => o.push(item));
            return enumerable(Enumerable.concat(o));
        },
        append(element) {
            return this.concat([element]);
        },
        group(selector, map = (a) => a) {
            return enumerable(Enumerable.group(this, selector, map));
        },
        lookup(parentSelector, selector, map = (elem) => elem) {
            return enumerable(Enumerable.lookup(this, parentSelector, selector, map));
        },
        min(selector) {
            return Enumerable.min(this, selector);
        },
        max(selector) {
            return Enumerable.max(this, selector);
        },
        toArray() {
            return Array.from(this);
        },
        toObject(keySelector, map = (elem) => elem) {
            return Enumerable.toObject(this, keySelector, map);
        },
        unwindMap(selector, mapSelector) {
            return enumerable(Enumerable.unwindMap(this, selector, mapSelector));
        },
        unwind(selector) {
            return enumerable(Enumerable.unwind(this, selector));
        },
        [Symbol.iterator]() {
            return iterable[Symbol.iterator]();
        }
    };
}
exports.enumerable = enumerable;
