"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.anyFirst = exports.arrayVisitor = void 0;
const expressions_1 = require("./expressions");
const guard_1 = require("./guard");
const enumerable_1 = require("../enumerable");
let deepEqual = function (left, right) {
    if (Array.isArray(left) && Array.isArray(right)) {
        return JSON.stringify(left) == JSON.stringify(right);
    }
    if (Array.isArray(left) && !Array.isArray(right)) {
        return left.indexOf(right) >= 0;
    }
    return left === right;
};
let withArraySome = function (left, right, fn) {
    if (Array.isArray(left) && !Array.isArray(right)) {
        return left.some(x => fn(x, right));
    }
    return fn(left, right);
};
let visitorBinary = function (arrayVisitor, source, binary) {
    let left = arrayVisitor(source).visit(binary.left);
    let right = arrayVisitor(source).visit(binary.right);
    return (function () {
        switch (binary.operator) {
            case expressions_1.Operator.eq:
                return deepEqual(left, right);
            case expressions_1.Operator.ne:
                return !deepEqual(left, right);
            case expressions_1.Operator.lt:
                return withArraySome(left, right, (a, b) => a < b);
            case expressions_1.Operator.lte:
                return withArraySome(left, right, (a, b) => a <= b);
            case expressions_1.Operator.gt:
                return withArraySome(left, right, (a, b) => a > b);
            case expressions_1.Operator.gte:
                return withArraySome(left, right, (a, b) => a >= b);
            case expressions_1.Operator.and:
                return left && right;
            case expressions_1.Operator.or:
                return left || right;
            default:
                throw new Error('binary operator could not found');
        }
    })();
};
let propertyVisitor = function (arrayVisitor, source, property) {
    let getValue = function (s, name) {
        if (s == null) {
            return null;
        }
        let sp = name.split('.');
        if (sp.length > 1) {
            let result = s[sp[0]];
            return getValue(result, sp.splice(1, sp.length - 1).join('.'));
        }
        return s[name];
    };
    return getValue(source, property.name);
};
let callVisitor = function (arrayVisitor, source, call) {
    let getArgs = function () { return (call.args || []).map(arg => arrayVisitor(source).visit(arg)); };
    if (call.method === expressions_1.CallExpressionType.exists) {
        let isExists = Object.keys(source).indexOf(call.target["name"]) >= 0;
        let check = getArgs()[0];
        return check ? isExists : !isExists;
    }
    let target = arrayVisitor(source).visit(call.target);
    if (target == null) {
        throw new Error('[target] is null for calling');
    }
    let applyStringMethod = function (value) {
        let args = getArgs();
        let arg = args[0];
        let _case = args[1] || 'g';
        value = value == null || _case === 'g' ? value : value.toLowerCase();
        arg = arg == null || _case === 'g' ? arg : arg.toLowerCase();
        switch (call.method) {
            case expressions_1.CallExpressionType.startsWith:
                return value.startsWith(arg);
            case expressions_1.CallExpressionType.endsWith:
                return value.endsWith(arg);
            case expressions_1.CallExpressionType.contains:
                return value.includes(arg);
            case expressions_1.CallExpressionType.equal:
                return value === arg;
            case expressions_1.CallExpressionType.in:
                return arg.includes(value);
            case expressions_1.CallExpressionType.size:
                return value.length;
            default:
                throw new Error(call.method + " could not found");
        }
    };
    let applyArrayMethod = function (value) {
        switch (call.method) {
            case expressions_1.CallExpressionType.all:
                let arg = getArgs()[0];
                return value.length != 0 && value.every(a => arg.includes(a));
            case expressions_1.CallExpressionType.elemMatch:
                return value.some(elem => arrayVisitor(elem).visit(call.args[0]));
            case expressions_1.CallExpressionType.size:
                return value.length;
            default:
                throw new Error(call.method + " could not found");
        }
    };
    let applyNumberMethod = function (value) {
        let args = getArgs();
        switch (call.method) {
            case expressions_1.CallExpressionType.in:
                return args[0].includes(value);
            default:
                throw new Error(call.method + " could not found");
        }
    };
    let methodFunction = function (t) {
        switch (typeof t) {
            case 'string':
                return applyStringMethod;
            case 'number':
                return applyNumberMethod;
            case 'bigint':
                return applyNumberMethod;
            default:
                break;
        }
        if (Array.isArray(t)) {
            return applyArrayMethod;
        }
        throw new Error('could not find apply function for ' + t);
    };
    return methodFunction(target)(target);
};
let callChainVisitor = function (arrayVisitor, source, callChain) {
    let sourceReturn = source;
    if (callChain.expression != null) {
        sourceReturn = arrayVisitor(source).visit(callChain.expression);
    }
    if (callChain.next == null) {
        return sourceReturn;
    }
    return arrayVisitor(sourceReturn).visit(callChain.next);
};
let sourceCallVisitor = function (arrayVisitor, source, sourceCall) {
    switch (sourceCall.method) {
        case expressions_1.SourceCallType.filter:
            return enumerable_1.iterate((function* () {
                for (let item of source) {
                    let result = arrayVisitor(item).visit(sourceCall.args[0]);
                    if (result) {
                        yield item;
                    }
                }
            }));
        case expressions_1.SourceCallType.first:
            return enumerable_1.iterate((function* () {
                for (let item of source) {
                    let result = arrayVisitor(item).visit(sourceCall.args[0]);
                    if (result) {
                        yield item;
                    }
                }
            }));
        case expressions_1.SourceCallType.sort:
            let order = function (a, b, type, key) {
                let aValue = arrayVisitor(a).visit({ type: expressions_1.ExpressionType.property, name: key });
                let bValue = arrayVisitor(b).visit({ type: expressions_1.ExpressionType.property, name: key });
                if (type === 'asc') {
                    return typeof aValue === 'string' ? aValue.localeCompare(bValue) : aValue - bValue;
                }
                return typeof bValue === 'string' ? bValue.localeCompare(aValue) : bValue - aValue;
            };
            let o = arrayVisitor(source).visit(sourceCall.args[0]);
            let result = source;
            Object.keys(o).forEach((key) => {
                if (key == null) {
                    throw new Error('sort type is null');
                }
                result = Array.from(source).sort((a, b) => order(a, b, o[key], key));
            });
            return enumerable_1.iterate(function* () {
                for (let item of result)
                    yield item;
            });
        case expressions_1.SourceCallType.skip:
            return enumerable_1.iterate(function* () {
                let i = 0;
                let skipCount = arrayVisitor(source).visit(sourceCall.args[0]);
                guard_1.Guard.isNumber(skipCount);
                for (let item of source) {
                    i++;
                    if (i > skipCount) {
                        yield item;
                    }
                }
            });
        case expressions_1.SourceCallType.take:
            let takeCount = arrayVisitor(source).visit(sourceCall.args[0]);
            guard_1.Guard.isNumber(takeCount);
            return enumerable_1.iterate(function* () {
                let i = 0;
                for (let item of source) {
                    if (i === takeCount) {
                        return;
                    }
                    yield item;
                    i++;
                }
            });
        case expressions_1.SourceCallType.count:
            return (function () {
                if (Array.isArray(source)) {
                    let count = source.length;
                    return enumerable_1.iterate(function* () {
                        yield {
                            count
                        };
                    });
                }
                let count = 0;
                for (let item of source)
                    count++;
                return enumerable_1.iterate(function* () {
                    yield {
                        count
                    };
                });
            })();
        case expressions_1.SourceCallType.select:
            let props = sourceCall.args.map(x => arrayVisitor(source).visit(x));
            let allProps = [{}].concat(props).reduce((r, n) => {
                if (typeof n === "object") {
                    Object.keys(n).forEach((key) => {
                        r[key] = 1;
                    });
                }
                else {
                    r[n] = 1;
                }
                return r;
            });
            let keys = Object.keys(allProps);
            return enumerable_1.iterate(function* () {
                for (let item of source) {
                    let o = {};
                    keys.forEach((key) => {
                        o[key] = item[key];
                    });
                    yield o;
                }
            });
        default:
            throw new Error(sourceCall.method + " could not find");
    }
};
function arrayVisitor(source) {
    let base = expressions_1.visitor();
    base.binary = function (binary) {
        return visitorBinary(arrayVisitor, source, binary);
    };
    base.property = function (property) {
        return propertyVisitor(arrayVisitor, source, property);
    };
    base.constant = function (constant) {
        return constant.value;
    };
    base.call = function (call) {
        return callVisitor(arrayVisitor, source, call);
    };
    base.callChain = function (callChain) {
        return callChainVisitor(arrayVisitor, source, callChain);
    };
    base.sourceCall = function (sourceCall) {
        return sourceCallVisitor(arrayVisitor, source, sourceCall);
    };
    return base;
}
exports.arrayVisitor = arrayVisitor;
function anyFirst(expression) {
    if (expression == null) {
        return false;
    }
    let base = expressions_1.visitor();
    let anyFirst = false;
    base.callChain = function (callChain) {
        let e = callChain.expression;
        base.visit(e);
        if (callChain.next) {
            base.visit(callChain.next);
        }
    };
    base.sourceCall = function (sourceCall) {
        if (sourceCall.method === expressions_1.SourceCallType.first) {
            anyFirst = true;
        }
        if (sourceCall.method === expressions_1.SourceCallType.count) {
            anyFirst = true;
        }
    };
    base.visit(expression);
    return anyFirst;
}
exports.anyFirst = anyFirst;
